/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

// File copied from the Thunderbird source tree. It was removed in Thunderbird 90.
// We still need it for compatibility with Thunderbird 68.
// This is a stripped-down version only providing fixIterator.

/**
 * This file contains helper methods for dealing with XPCOM iterators (arrays
 * and enumerators) in JS-friendly ways.
 */

const EXPORTED_SYMBOLS = ["fixIterator"];

var JS_HAS_SYMBOLS = typeof Symbol === "function";
var ITERATOR_SYMBOL = JS_HAS_SYMBOLS ? Symbol.iterator : "@@iterator";

/**
 * Given a JS array, JS iterator, or one of a variety of XPCOM collections or
 * iterators, return a JS iterator suitable for use in a for...of expression.
 *
 * Currently, we support the following types of XPCOM iterators:
 *   nsIArray
 *   nsISimpleEnumerator
 *
 * Note that old-style JS iterators are explicitly not supported in this
 * method, as they are going away.
 *
 *   @param aEnum  the enumerator to convert
 *   @param aIface (optional) an interface to QI each object to prior to
 *                 returning
 *
 *   @note This returns an object that can be used in 'for...of' loops.
 *         Do not use 'for each...in' or 'for...in'.
 *         This does *not* return an Array object. To create such an array, use
 *         let array = toArray(fixIterator(xpcomEnumerator));
 */
function fixIterator(aEnum, aIface) {
  // If the input is an array, nsISimpleEnumerator or something that sports Symbol.iterator,
  // then the original input is sufficient to directly return. However, if we want
  // to support the aIface parameter, we need to do a lazy version of
  // Array.prototype.map.
  if (
    Array.isArray(aEnum) ||
    aEnum instanceof Ci.nsISimpleEnumerator ||
    ITERATOR_SYMBOL in aEnum
  ) {
    if (!aIface) {
      return aEnum[ITERATOR_SYMBOL]();
    }
    return (function*() {
      for (let o of aEnum) {
        yield o.QueryInterface(aIface);
      }
    })();
  }

  let face = aIface || Ci.nsISupports;
  // Figure out which kind of array object we have.
  // First try nsIArray (covers nsIMutableArray too).
  if (aEnum instanceof Ci.nsIArray) {
    return (function*() {
      let count = aEnum.length;
      for (let i = 0; i < count; i++) {
        yield aEnum.queryElementAt(i, face);
      }
    })();
  }

  // We got something unexpected, notify the caller loudly.
  throw new Error(
    "An unsupported object sent to fixIterator: " +
      ("toString" in aEnum ? aEnum.toString() : aEnum)
  );
}
